import React, { memo, useRef, useEffect, useCallback } from 'react';
import { useDeepCompareEffect } from 'use-deep-compare';
import { Options, Plot } from '@antv/g2plot';
import { Axis } from '@antv/g2plot/lib/types/axis';
import { Tooltip } from '@antv/g2plot/lib/types/tooltip';
import { ChartEmpty } from './components';

export type Data = Record<string, any>[];

export type ChartOptions = Omit<Options, 'data'> & {
  data?: Record<string, any>[];
  [key: string]: any;
};

export type ExtendFormatter = {
  formatterStr?: string;
};

export type ExtendAxis = Axis & {
  label?: {
    formatterStr?: string;
  };
};

// 序列类型图表的通用类型
export type SeqChartOptions<T extends Options> = Omit<
  T,
  'data' | 'tooltip' | 'xAxis' | 'yAxis' | 'xField' | 'yField'
> & {
  data?: Data;
  tooltip?: Tooltip & ExtendFormatter;
  xAxis?: ExtendAxis;
  yAxis?: ExtendAxis;
  xField?: string;
  yField?: string;
  yFields?: Array<string>;
  legendPosition?: string;
  legendSubPosition?: string;
  title?: string;
  subTitle?: string;
};

export type CreateChartOptions<T> = {
  transformData?: (options: T) => Data;
  transformConfig?: (options: T) => ChartOptions;
};

// 这里不太好解决 Plot<T> 类泛型问题，先忽略掉类型
export type ChartType = any;

export const createChart = <TOptions extends ChartOptions>(
  Chart: ChartType,
  { transformData, transformConfig }: CreateChartOptions<TOptions>,
) =>
  memo((props: TOptions) => {
    const { data, onElementClick, ...config } = props;
    const chartRef = useRef<Plot<TOptions> | null>(null);
    const chartContainerRef = useRef<HTMLDivElement | null>(null);
    const propsRef = useRef<TOptions>(props);
    propsRef.current = props;
    const handleElementClick = useCallback(
      event => {
        onElementClick(event.data);
      },
      [onElementClick],
    );
    // 图表实例初始化
    useEffect(() => {
      if (!chartContainerRef.current) {
        return () => null;
      }

      const chartConfig =
        typeof transformConfig === 'function'
          ? transformConfig(propsRef.current)
          : config;
      const chartData =
        typeof transformData === 'function'
          ? transformData(propsRef.current)
          : data;

      chartRef.current = new Chart(chartContainerRef.current, {
        ...chartConfig,
        data: chartData,
      });

      if (propsRef.current.onElementClick) {
        chartRef.current?.on('element:click', handleElementClick);
      }

      return () => {
        chartRef.current?.off('element:click', handleElementClick);
        chartRef.current?.destroy();
      };
    }, [chartContainerRef.current, handleElementClick]);

    // 更新图表数据、配置
    useDeepCompareEffect(() => {
      // TODO 优化点：第一次渲染数据为空不渲染，之后数据为空需渲染为空
      if (!chartRef.current || !data) {
        return;
      }

      try {
        const chartData =
          typeof transformData === 'function'
            ? transformData(propsRef.current)
            : data;
        const chartConfig = (
          typeof transformConfig === 'function'
            ? transformConfig(propsRef.current)
            : config
        ) as TOptions;

        window.requestIdleCallback(() => {
          chartRef.current?.update({ ...chartConfig, data: chartData });
        });
      } catch {
        console.error('chart update found error');
      }
    }, [data, config]);

    const Title = (
      <div className="title" style={{ marginBottom: '20px' }}>
        <h1 style={{ marginBottom: '-10px' }}>{config.title}</h1>
        <h2 style={{ color: 'rgb(78, 89, 105)' }}>{config.subTitle}</h2>
      </div>
    );

    return (
      <div className="relative">
        {Title}
        <div ref={chartContainerRef} />
        {!data?.length ? <ChartEmpty>暂无数据</ChartEmpty> : null}
      </div>
    );
  });
